@@ -224,13 +224,28 @@ private:
     void startUpdating();
 
     /**
-     * Creates previews for the items starting from the first item of the
-     * given list.
+     * Loads the icons for the visible items. After 200 ms, the function
+     * stops determining mime types and only loads preliminary icons.
+     * This is a compromise that prevents that
+     * (a) the GUI is blocked for more than 200 ms, and
+     * (b) "unknown" icons could be shown in the view.
+     */
+    void updateVisibleIcons();
+
+    /**
+     * Tries to load at least preliminary icons (without determining the
+     * mime type) for all items for \a timeout milliseconds.
+     */
+    void updateAllIconsFast(int timeout);
+
+    /**
+     * Creates previews for the items starting from the first item in
+     * m_pendingPreviewItems.
      * @see slotGotPreview()
      * @see slotPreviewFailed()
      * @see slotPreviewJobFinished()
      */
-    void startPreviewJob(const KFileItemList items);
+    void startPreviewJob();
 
     /**
      * Ensures that icons, previews, and other roles are determined for any
@@ -311,10 +326,10 @@ private:
     // Items for which the sort role still has to be determined.
     QSet<KFileItem> m_pendingSortRoleItems;
 
-    // While the sort role is being resolved, we also keep the indexes
-    // in a sorted list. The reason is that this enables us to determine
-    // both the sort role and the icon for the visible items first.
-    QList<int> m_pendingSortRoleIndexes;
+    // Determines if the next call of startUpdating() will try to do a fast
+    // icon loading (i.e., without determining the mime type) for all items.
+    bool m_hasUnknownIcons;
+    int m_firstIndexWithoutIcon;
 
     // Indexes of items which still have to be handled by
     // resolveNextPendingRoles().
